import typing as t
from inspect import get_annotations, signature

from tramp.optionals import Optional

import bevy.injections as injections
import bevy.registries as registries
from bevy.context_vars import GlobalContextMixin, get_global_container, global_container
from bevy.dependencies import DependencyMetadata
from bevy.hooks import Hook

type Instance = t.Any


class Container(GlobalContextMixin, var=global_container):
    """Stores instances for dependencies and provides utilities for injecting dependencies into callables at runtime.

    Containers can be branched to isolate dependencies while still sharing the same registry and preexisting instances."""
    def __init__(self, registry: "registries.Registry", *, parent: "Container | None" = None):
        super().__init__()
        self.registry = registry
        self.instances: dict[t.Type[Instance], Instance] = {}
        self._parent = parent

    def branch(self) -> "Container":
        """Creates a branch off of the current container, isolating dependencies from the parent container. Dependencies
        existing on the parent container will be shared with the branched container."""
        return Container(registry=self.registry, parent=self)

    def call[**P, R](
        self, func: "t.Callable[P, R] | injections.InjectionFunctionWrapper[P, R]", *args: P.args, **kwargs: P.kwargs
    ) -> R:
        """Calls a function or class with the provided arguments and keyword arguments, injecting dependencies from the
        container. This will create instances of any dependencies that are not already stored in the container or its
        parents."""
        match func:
            case injections.InjectionFunctionWrapper() as wrapper:
                return wrapper.call_using(self, *args, **kwargs)

            case _:
                return self._call(func, *args, **kwargs)

    @t.overload
    def get[T: Instance](self, dependency: t.Type[T]) -> T:
        ...

    @t.overload
    def get[T: Instance, D](self, dependency: t.Type[T], *, default: D) -> T | D:
        ...

    def get[T: Instance, D](self, dependency: t.Type[T], **kwargs) -> T | D:
        """Gets an instance of the desired dependency from the container. If the dependency is not already stored in the
        container or its parents, a new instance will be created and stored for reuse. When a default is given it will
        be returned instead of creating a new instance, the default is not stored."""
        using_default = False
        match self.registry.hooks[Hook.GET_INSTANCE].handle(self, dependency):
            case Optional.Some(v):
                instance = v

            case Optional.Nothing():
                if dep := self._get_existing_instance(dependency):
                    instance = dep.value
                else:
                    dep = None
                    if self._parent:
                        dep = self._parent.get(dependency, default=None)

                    if dep is None:
                        if "default" in kwargs:
                            dep = kwargs["default"]
                            using_default = True
                        else:
                            dep = self._create_instance(dependency)

                    instance = dep

            case _:
                raise ValueError(f"Invalid value for dependency: {dependency}, must be an Optional type.")

        instance = self.registry.hooks[Hook.GOT_INSTANCE].filter(self, instance)
        if not using_default:
            self.instances[dependency] = instance

        return instance

    def _call[**P, R](self, func: t.Callable[P, R] | t.Type[R], *args: P.args, **kwargs: P.kwargs) -> R:
        match func:
            case type():
                return self._call_type(func, *args, **kwargs)

            case _:
                return self._call_function(func, *args, **kwargs)

    def _call_function[**P, R](self, func: t.Callable[P, R], *args: P.args, **kwargs: P.kwargs) -> R:
        f = _unwrap_function(func)
        sig = signature(f)
        ns = f.__globals__
        annotations = get_annotations(f, globals=ns, eval_str=True)

        params = sig.bind_partial(*args, **kwargs)
        params.arguments |= {
            name: self.get(annotations[name]) if parameter.default.factory is None else parameter.default.factory(self)
            for name, parameter in sig.parameters.items()
            if isinstance(parameter.default, DependencyMetadata) and name not in params.arguments
        }
        return func(*params.args, **params.kwargs)

    def _call_type[T](self, type_: t.Type[T], *args, **kwargs) -> T:
        instance = type_.__new__(type_, *args, **kwargs)
        self.call(instance.__init__, *args, **kwargs)
        return instance

    def _create_instance(self, dependency: t.Type[Instance]) -> Instance:
        match self.registry.hooks[Hook.CREATE_INSTANCE].handle(self, dependency):
            case Optional.Some(v):
                instance = v

            case Optional.Nothing():
                for factory_type, factory in self.registry.factories.items():
                    if issubclass(dependency, factory_type):
                        instance = factory(self)
                        break

                else:
                    match self.registry.hooks[Hook.HANDLE_UNSUPPORTED_DEPENDENCY].handle(self, dependency):
                        case Optional.Some(v):
                            instance = v

                        case Optional.Nothing():
                            raise TypeError(f"No value found for {dependency}")

                        case _:
                            raise ValueError(f"Invalid value for dependency: {dependency}, must be an Optional type.")

            case _:
                raise ValueError(f"Invalid value for dependency: {dependency}, must be an Optional type.")

        return self.registry.hooks[Hook.CREATED_INSTANCE].filter(self, instance)

    def _get_existing_instance(self, dependency: t.Type[Instance]) -> Optional[Instance]:
        for instance_type, instance in self.instances.items():
            if issubclass(dependency, instance_type):
                return Optional.Some(instance)

        return Optional.Nothing()


def _unwrap_function(func):
    if hasattr(func, "__wrapped__"):
        return _unwrap_function(func.__wrapped__)

    if hasattr(func, "__func__"):
        return _unwrap_function(func.__func__)

    return func


@t.overload
def get_container() -> Container:
    ...


@t.overload
def get_container(obj: Container | None) -> Container:
    ...


@t.overload
def get_container(*, using_registry: "registries.Registry") -> Container:
    ...


@t.overload
def get_container(obj: Container | None, *, using_registry: "registries.Registry") -> Container:
    ...


def get_container(*args, **kwargs) -> Container:
    """Gets a container from the global context, a provided container, or a registry.

    With no parameters it returns the global container. Note that this will create a new container if the global context
    does not already have one.

    When given a container it returns that container unless it is None. When it is None it returns the global container.
    Note that this will create a new container if the provided container is None and the global context does not already
    have one.

    A registry can be passed to use in place of the global registry with the using_registry keyword argument. Containers
    created by this registry will not be stored in the global context."""
    match args:
        case (Container() as container,):
            return container

        case () | (None,):
            match kwargs:
                case {"using_registry": registries.Registry() as registry}:
                    return registry.create_container()

                case {}:
                    return get_global_container()

                case _:
                    names = (name for name in kwargs if name not in {"using_registry"})
                    raise NameError(
                        f"Invalid keyword name(s): {', '.join(names)}"
                    )

        case _:
            raise ValueError(f"Invalid positional arguments: {args}\nExpected no parameters, a Container, or None.")



import typing as t
from functools import update_wrapper

import bevy.containers as containers


class _DescriptorProtocol(t.Protocol):
    def __get__(self, instance, owner) -> t.Callable:
        ...

    def __set__(self, instance, value):
        ...


class InjectionFunctionWrapper[**P, R]:
    """Wraps a callable to allow for injection of dependencies when called using a container. Called directly this uses
    the global container. A different container can be used with the `call_using` method."""
    def __init__(self, func: t.Callable[P, R]):
        self._func = func
        update_wrapper(self, func)

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        return self.call_using(containers.get_container(), *args, **kwargs)

    def __get__(self, instance, owner):
        return InjectionFunctionWrapper(t.cast(_DescriptorProtocol, self._func).__get__(instance, owner))

    def call_using(self, container: "containers.Container", *args: P.args, **kwargs: P.kwargs) -> R:
        """Calls the wrapped function using the provided container for dependency injection."""
        return container.call(self._func, *args, **kwargs)


def inject[**P, R](func: t.Callable[P, R]) -> InjectionFunctionWrapper[P, R]:
    """Simple wrapper around InjectionFunctionWrapper to make it cleaner to use."""
    return InjectionFunctionWrapper(func)



import os
from contextvars import ContextVar

import bevy.containers as c
import bevy.registries as r

global_container: "ContextVar[c.Container]" = ContextVar("global_container")
global_registry: "ContextVar[r.Registry]" = ContextVar("global_registry")


class GlobalContextDisabledError(Exception):
    """Raised when the global context is disabled by the BEVY_ENABLE_GLOBAL_CONTEXT environment variable."""


yes_no_mapping = {
    "yes": True,
    "no": False,
    "true": True,
    "false": False,
    "1": True,
    "0": False,
    "y": True,
    "n": False,
}


def is_global_context_enabled() -> bool:
    """Returns True if the global context is enabled, False otherwise."""
    var = os.getenv("BEVY_ENABLE_GLOBAL_CONTEXT", "yes")
    return yes_no_mapping.get(var.casefold(), True)


def get_global_registry() -> "r.Registry":
    """Gets the global registry. If no registry exists, creates a new one. Raises GlobalContextNotAllowedError if the
    BEVY_ALLOW_GLOBAL_CONTEXT environment variable is set to False."""
    if not is_global_context_enabled():
        raise GlobalContextDisabledError("Global context is disabled. You must provide a registry to use.")

    try:
        registry = global_registry.get()
    except LookupError:
        global_registry.set(
            registry := r.Registry()
        )

    return registry


def get_global_container() -> "c.Container":
    """Gets the global container. If no container exists, creates a new one using the global registry. Raises
    GlobalContextNotAllowedError if the BEVY_ALLOW_GLOBAL_CONTEXT environment variable is set to False."""
    if not is_global_context_enabled():
        raise GlobalContextDisabledError("Global context is disabled. You must provide a container to use.")

    try:
        container = global_container.get()
    except LookupError:
        global_container.set(
            container := get_global_registry().create_container()
        )

    return container


class GlobalContextMixin:
    """This mixin allows instances to be loaded into a predefined contextvar using a context manager."""
    def __init_subclass__(cls, *, var: ContextVar, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._context_var = var

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._reset_tokens = []

    def __enter__(self):
        self._reset_tokens.append(self._context_var.set(self))
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._context_var.reset(self._reset_tokens.pop())



import functools
from enum import Enum
from typing import Any, Callable, TYPE_CHECKING

from tramp.optionals import Optional

import bevy.registries as r

if TYPE_CHECKING:
    from bevy.containers import Container

type HookFunction[T] = "Callable[[Container, Type[T]], T]"


class Hook(Enum):
    GET_INSTANCE = "get_instance"
    GOT_INSTANCE = "got_instance"
    CREATE_INSTANCE = "create_instance"
    CREATED_INSTANCE = "created_instance"
    HANDLE_UNSUPPORTED_DEPENDENCY = "handle_unsupported_dependency"


class HookManager:
    """A utility type that makes it easier to work with collections of functions waiting for the
    hook to be triggered."""
    def __init__(self):
        self.callbacks = set()

    def add_callback(self, hook: HookFunction):
        """Adds a function that will be called when the hook is triggered."""
        self.callbacks.add(hook)

    def handle[T](self, container: "Container", value: T) -> Optional[Any]:
        """Iterates each callback and returns the first result."""
        for callback in self.callbacks:
            match callback(container, value):
                case Optional.Some() as v:
                    return v

        return Optional.Nothing()

    def filter[T](self, container: "Container", value: T) -> T:
        """Iterates all callbacks and updates the value when a callback returns a Some result."""
        for callback in self.callbacks:
            match callback(container, value):
                case Optional.Some(v):
                    value = v
                case Optional.Nothing():
                    pass

        return value


class HookWrapper[**P, R]:
    """Wraps a hook callback function to make it easier to register with a registry."""
    __match_args__ = ("hook_type",)

    def __init__(self, hook_type: Hook, func: Callable[P, R]):
        self.hook_type = hook_type
        self.func = func

        functools.update_wrapper(self, func)

    def __call__(self, container: "Container", value: P) -> Optional[R]:
        return self.func(container, value)

    def register_hook(self, registry: "r.Registry | None" = None):
        """Adds the callback to a registry for the hook type."""
        registry = r.get_registry(registry)
        registry.add_hook(self)


class _HookDecoratorDescriptor:
    def __init__(self):
        self.hook_type: Optional[Hook] = Optional.Nothing()

    def __get__(self, instance, owner):
        match self.hook_type:
            case Optional.Some(hook_type):
                return HookDecorator(hook_type)

            case Optional.Nothing():
                raise ValueError("Hook type is not yet set. Accessed before owning class definition fully created.")

            case _:
                raise ValueError("Invalid value for hook type.")

    def __set_name__(self, owner, name):
        self.hook_type = Optional.Some(Hook[name])


class HookDecorator[**P, R]:
    """A decorator that wraps a function in a hook type to simplifying adding to a registry. This class is aliased as
    "hooks" for convenience. It provides decorators for each hook type for even simpler syntax.

    Example:
        @hooks.GET_INSTANCE
        def foobar(container: Container, some_thing: Thing) -> Thing:
            ...
    """
    GET_INSTANCE = _HookDecoratorDescriptor()
    GOT_INSTANCE = _HookDecoratorDescriptor()
    CREATE_INSTANCE = _HookDecoratorDescriptor()
    CREATED_INSTANCE = _HookDecoratorDescriptor()
    HANDLE_UNSUPPORTED_DEPENDENCY = _HookDecoratorDescriptor()

    def __init__(self, hook_type: Hook):
        self.hook_type = hook_type

    def __call__(self, func: Callable[P, R]) -> HookWrapper[P, R]:
        return HookWrapper(self.hook_type, func)


hooks = HookDecorator



from collections.abc import Callable
from functools import update_wrapper
from typing import overload, Sequence, Type

import bevy.registries as r


class Factory[**P, T]:
    """A wrapper for dependency factories. This makes it easier to define factories that can handle various types and
    add them to the registry."""
    def __init__(self, dependency_types: Sequence[Type[T]], factory: "r.DependencyFactory[P, T]"):
        self.dependency_types = dependency_types
        self.factory = factory

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        return self.factory(*args, **kwargs)

    @overload
    def register_factory(self):
        ...

    @overload
    def register_factory(self, registry: "r.Registry | None"):
        ...

    def register_factory(self, registry: "r.Registry | None" = None):
        """Adds the factory to the registry for each dependency type the factory supports. If the registry is not
        provided or is None the global registry will be used."""
        registry = r.get_registry(registry)
        for dependency_type in self.dependency_types:
            registry.add_factory(self, dependency_type)


def factory[**P, T](*dependency_types: Type[T]) -> "Callable[[r.DependencyFactory[P, T]], r.DependencyFactory[P, T]]":
    """Decorator that wraps a factory function in a factory wrapper."""
    def decorator(dependency_factory: Factory[P, T]) -> r.DependencyFactory[P, T]:
        wrapper = Factory[P, T](dependency_types, dependency_factory)
        wrapper = update_wrapper(wrapper, dependency_factory)
        return wrapper

    return decorator


def create_type_factory[T](dependency_type: Type[T], *args, **kwargs) -> Factory[None, T]:
    """Creates a factory for a type that can be added to a registry. It takes the type and any arguments that should be
    passed to the type constructor."""
    return Factory([dependency_type], lambda _: dependency_type(*args, **kwargs))



from typing import Any, Callable, overload, TYPE_CHECKING

if TYPE_CHECKING:
    from bevy.containers import Container
    from bevy.factories import Factory


class DependencyMetadata:
    """Dependency metadata wrapper. A factory function can be provided for creating the dependency, this factory
    function must take a container instance as a parameter."""
    def __init__(self, factory: "Callable[[Container], Any] | None" = None):
        self.factory = factory


@overload
def dependency() -> Any:
    ...


@overload
def dependency[T](factory: "Callable[[Container], T] | Factory[None, T]") -> T:
    ...


def dependency[T](factory: "Callable[[Container], T] | Factory[None, T] | None" = None) -> T | Any:
    """Intended to be called as a parameter default value. It assigns the default to a dependency metadata object. A
    factory function can be passed that returns an instance of the dependency type."""
    return DependencyMetadata(factory)



from typing import Type

from tramp.optionals import Optional

from bevy.containers import Container
from bevy.hooks import hooks


@hooks.HANDLE_UNSUPPORTED_DEPENDENCY
def type_factory[T](container: Container, dependency: Type[T]) -> Optional[T]:
    """This hook adds a dependency factory that injects an instances of the dependency type."""
    if isinstance(dependency, type):
        return Optional.Some(container.call(dependency))

    return Optional.Nothing()



from collections import defaultdict
from typing import overload, Type

import bevy.containers as containers
import bevy.hooks as hooks
from bevy.context_vars import get_global_registry, global_registry, GlobalContextMixin
from bevy.factories import Factory

type DependencyFactory[T] = "Callable[[containers.Container], T]"


class Registry(GlobalContextMixin, var=global_registry):
    """Registries hold factories and hooks for creating and managing instances of objects. Containers are created from
    registries, and containers are used to create and cache instances of objects."""
    def __init__(self):
        super().__init__()
        self.hooks: dict[hooks.Hook, hooks.HookManager] = defaultdict(hooks.HookManager)
        self.factories: "dict[Type[containers.Instance], DependencyFactory[containers.Instance]]" = {}

    @overload
    def add_factory(self, factory: "DependencyFactory[containers.Instance]", for_type: "Type[containers.Instance]"):
        ...

    @overload
    def add_factory(self, factory: "Factory"):
        ...

    def add_factory(self, *args):
        """Adds a factory to the registry. Factories are used to create instances of objects. If an instance of Factory
        is passed its register_factory method is called to register it with the registry. If a callable and type is
        passed, the callable is stored as a factory for the type."""
        match args:
            case [Factory() as factory]:
                factory.register_factory(self)

            case [factory, type() as for_type] if callable(factory):
                self.factories[for_type] = factory

            case _:
                raise ValueError(f"Unexpected arguments to add_factory: {args}")

    @overload
    def add_hook(self, hook: "hooks.HookWrapper"):
        ...

    @overload
    def add_hook(self, hook_type: "hooks.Hook", func: "hooks.HookFunction"):
        ...

    def add_hook(self, *args):
        """Adds a callback to a hook. If a HookWrapper is passed, the hook is added to the registry. If a Hook type and
        a callable are passed, the callable is added as a callback to the hook."""
        match args:
            case [hooks.Hook() as hook_type, func] if callable(func):
                self.hooks[hook_type].add_callback(func)

            case [hooks.HookWrapper(hook_type) as hook]:
                self.hooks[hook_type].add_callback(hook)

            case _:
                raise ValueError(f"Unexpected arguments to add_hook: {args}")


    def create_container(self) -> "containers.Container":
        """Creates a new container bound to the registry."""
        return containers.Container(self)


@overload
def get_registry(registry: Registry | None) -> Registry:
    ...


@overload
def get_registry() -> Registry:
    ...


def get_registry(*args) -> Registry:
    """Returns a registry. If a registry is passed, it is returned. If no registry is passed or None is passed, the
    global registry is returned. This creates a new global registry if it is needed and doesn't already exist."""
    match args:
        case [Registry() as registry]:
            return registry

        case [None]:
            return get_global_registry()

        case []:
            return get_global_registry()

        case _:
            raise ValueError(f"Unexpected arguments to get_registry: {args}")


